#pragma once

#include "async.h"
#include "io.h"
#include "message.h"

namespace saw {

template <typename Codec, typename Incoming, typename Outgoing,
		  typename InContainer = MessageContainer<Incoming>,
		  typename OutContainer = MessageContainer<Outgoing>,
		  typename BufferT = ring_buffer>
class StreamingIoPeer {
public:
	/**
	 *
	 */
	StreamingIoPeer(
		own<conveyor_feeder<HeapMessageRoot<Incoming, InContainer>>> feed,
		own<AsyncIoStream> stream, Codec codec, BufferT in, BufferT out);
	/**
	 *
	 */
	StreamingIoPeer(
		own<conveyor_feeder<HeapMessageRoot<Incoming, InContainer>>> feed,
		own<AsyncIoStream> stream);

	/**
	 * Deleted copy and move constructors
	 */
	SAW_FORBID_COPY(StreamingIoPeer);
	SAW_FORBID_MOVE(StreamingIoPeer);

	/**
	 * Send a message to the remote peer
	 */
	void send(HeapMessageRoot<Outgoing, OutContainer> builder);

	/**
	 * A phantom conveyor feeder. Meant for interfacing with other components
	 */
	conveyor_feeder<HeapMessageRoot<Outgoing, OutContainer>> &feeder();

	conveyor<void> onReadDisconnected();

private:
	class Peerconveyor_feeder final
		: public conveyor_feeder<HeapMessageRoot<Outgoing, OutContainer>> {
	public:
		Peerconveyor_feeder(
			StreamingIoPeer<Codec, Incoming, Outgoing, InContainer,
							OutContainer, BufferT> &peer_)
			: peer{peer_} {}

		void feed(T &&data) override { (void)data; }

		void fail(error &&error) override { (void)error; }

		size_t space() const override { return 0; }

		size_t queued() const override { return 0; }

	private:
		StreamingIoPeer<Codec, Incoming, Outgoing, InContainer, OutContainer,
						BufferT> &peer;
	};

private:
	own<conveyor_feeder<HeapMessageRoot<Incoming, InContainer>>>
		incoming_feeder = nullptr;

	own<AsyncIoStream> io_stream;

	Codec codec;

	BufferT in_buffer;
	BufferT out_buffer;

	sink_conveyor sink_read;
	sink_conveyor sink_write;

	Peerconveyor_feeder conveyor_feeder;
};

/**
 * Setup new streaming io peer with the provided network protocols.
 * This is a convenience wrapper intended for a faster setup of this class
 */
template <typename Codec, typename Incoming, typename Outgoing,
		  typename InContainer = MessageContainer<Incoming>,
		  typename OutContainer = MessageContainer<Outgoing>,
		  typename BufferT = ring_buffer>
std::pair<own<StreamingIoPeer<Codec, Incoming, Outgoing, InContainer,
							  OutContainer, BufferT>>,
		  conveyor<HeapMessageRoot<Incoming, InContainer>>>
newStreamingIoPeer(own<AsyncIoStream> stream);

} // namespace saw

#include "io_peer.tmpl.h"
