namespace saw {

template <typename Codec, typename Incoming, typename Outgoing,
		  typename InContainer, typename OutContainer, typename BufferT>
StreamingIoPeer<Codec, Incoming, Outgoing, InContainer, OutContainer, BufferT>::
	StreamingIoPeer(
		own<ConveyorFeeder<HeapMessageRoot<Incoming, InContainer>>> feed,
		own<AsyncIoStream> str)
	: StreamingIoPeer{std::move(feed), std::move(str), {}, {}, {}} {}

template <typename Codec, typename Incoming, typename Outgoing,
		  typename InContainer, typename OutContainer, typename BufferT>
StreamingIoPeer<Codec, Incoming, Outgoing, InContainer, OutContainer, BufferT>::
	StreamingIoPeer(
		own<ConveyorFeeder<HeapMessageRoot<Incoming, InContainer>>> feed_,
		own<AsyncIoStream> stream_, Codec codec_, BufferT in_, BufferT out_)
	: incoming_feeder{std::move(feed_)}, io_stream{std::move(stream_)},
	  codec{std::move(codec_)}, in_buffer{std::move(in_)}, out_buffer{std::move(
															   out_)},
	  sink_read{io_stream->readDone()
					.then([this](size_t bytes) -> error_or<void> {
						in_buffer.writeAdvance(bytes);

						if (in_buffer.writeSegmentLength() == 0) {
							return critical_error("Message too long");
						}

						io_stream->read(&in_buffer.write(), 1,
										in_buffer.writeSegmentLength());

						while (true) {
							auto root =
								heapMessageRoot<Incoming, InContainer>();
							auto builder = root.build();

							error error =
								codec.template decode<Incoming, InContainer>(
									builder, in_buffer);
							if (error.is_critical()) {
								return error;
							}

							if (!error.failed()) {
								incoming_feeder->feed(std::move(root));
							} else {
								break;
							}
						}

						return Void{};
					})
					.sink([this](error error) {
						incoming_feeder->fail(error.copy_error());

						return error;
					})},
	  sink_write{io_stream->writeDone()
					 .then([this](size_t bytes) -> error_or<void> {
						 out_buffer.readAdvance(bytes);
						 if (out_buffer.readCompositeLength() > 0) {
							 io_stream->write(&out_buffer.read(),
											  out_buffer.readSegmentLength());
						 }

						 return Void{};
					 })
					 .sink()} {
	io_stream->read(&in_buffer.write(), 1, in_buffer.writeSegmentLength());
}

template <typename Codec, typename Incoming, typename Outgoing,
		  typename InContainer, typename OutContainer, typename BufferT>
void StreamingIoPeer<Codec, Incoming, Outgoing, InContainer, OutContainer,
					 BufferT>::send(HeapMessageRoot<Outgoing, OutContainer>
										msg) {
	bool restart_write = out_buffer.readSegmentLength() == 0;

	error err =
		codec.template encode<Outgoing, OutContainer>(msg.read(), out_buffer);
	if (err.failed()) {
		return err;
	}

	if (restart_write) {
		io_stream->write(&out_buffer.read(), out_buffer.readSegmentLength());
	}

	return no_error();
}

template <typename Codec, typename Incoming, typename Outgoing,
		  typename InContainer, typename OutContainer, typename BufferT>
Conveyor<void> StreamingIoPeer<Codec, Incoming, Outgoing, InContainer,
							   OutContainer, BufferT>::onReadDisconnected() {
	return io_stream->onReadDisconnected();
}

template <typename Codec, typename Incoming, typename Outgoing,
		  typename InContainer, typename OutContainer, typename BufferT>
std::pair<own<StreamingIoPeer<Codec, Incoming, Outgoing, InContainer,
							  OutContainer, BufferT>>,
		  Conveyor<HeapMessageRoot<Incoming, InContainer>>>
newStreamingIoPeer(own<AsyncIoStream> stream) {
	auto caf = newConveyorAndFeeder<HeapMessageRoot<Incoming, InContainer>>();

	return {heap<StreamingIoPeer<Codec, Incoming, Outgoing, InContainer,
								 OutContainer, BufferT>>(std::move(caf.feeder),
														 std::move(stream)),
			std::move(caf.conveyor)};
}

} // namespace saw
