#pragma once

#include "../common.h"
#include "../io.h"

#include <optional>
#include <variant>

namespace saw {
class Tls;

class TlsServer final : public Server {
private:
	own<Server> internal;

public:
	TlsServer(own<Server> srv);

	conveyor<own<IoStream>> accept() override;
};

class TlsNetwork final : public Network {
private:
	Tls& tls;
	Network &internal;
public:
	TlsNetwork(Tls& tls_, Network &network_);

	conveyor<own<NetworkAddress>> resolveAddress(const std::string &addr, uint16_t port = 0) override;
	
	own<Server> listen(NetworkAddress& address) override;

	conveyor<own<IoStream>> connect(NetworkAddress& address) override;

	own<Datagram> datagram(NetworkAddress& address) override;
};

/**
* Tls context class.
* Provides tls network class which ensures the usage of tls encrypted connections
*/
class Tls {
private:
	class Impl;
	own<Impl> impl;
public:
	Tls();
	~Tls();

	struct Version {
		struct Tls_1_0{};
		struct Tls_1_1{};
		struct Tls_1_2{};
	};

	struct Options {
	public:
		Version version;
	};

	Network& tlsNetwork();

	Impl &getImpl();
private:
	Options options;
};

std::optional<own<TlsNetwork>> setupTlsNetwork(Network &network);

} // namespace saw
